/*
 *   Creation Date: <2010/04/02 12:00:00 mcayland>
 *   Time-stamp: <2010/04/02 12:00:00 mcayland>
 *
 *	<openbios.h>
 *
 *	General OpenBIOS initialization
 *
 *   Copyright (C) 2010 Mark Cave-Ayland (mark.cave-ayland@siriusit.co.uk)
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   version 2
 *
 */

#ifndef _H_LIBOPENBIOS
#define _H_LIBOPENBIOS

extern void	openbios_init( void );

#endif   /* _H_LIBOPENBIOS */
