/*
 *   Creation Date: <2004/08/28 18:38:22 greg>
 *   Time-stamp: <2004/08/28 18:38:22 greg>
 *
 *	<tree.c>
 *
 *	device tree setup
 *
 *   Copyright (C) 2004 Greg Watson
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   version 2
 *
 */

#include "config.h"
#include "libopenbios/bindings.h"

void devtree_init( void )
{
	fword("init-briq-tree");
}
