/*
 *   Creation Date: <2004/08/28 18:03:25 stepan>
 *   Time-stamp: <2004/08/28 18:03:25 stepan>
 *
 *	<briq/kernel.c>
 *
 *   Copyright (C) 2004 Stefan Reinauer
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   version 2
 *
 */

#include "briq-dict.h"
#include "../kernel.c"
